import logging
import sys
from ._version import __version__

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

def main():
    log.info(f"Neural ADMIXTURE - Version {__version__}")
    log.info("[CHANGELOG] Default P initialization was changed to 'pckmeans' in version 1.3.0.")
    log.info("[CHANGELOG] Warmup training for initialization of Q was added in version 1.3.0 to improve training stability (only for `pckmeans`).")
    log.info("[CHANGELOG] Convergence check changed so it is performed after 15 epochs in version 1.3.0 to improve training stability.")
    log.info("[CHANGELOG] Default learning rate was changed to 1e-5 instead of 1e-4 in version 1.3.0 to improve training stability.")
    arg_list = tuple(sys.argv)
    assert len(arg_list) > 1, 'Please provide either the argument "train" or "infer" to choose running mode.'
    if sys.argv[1] == 'train':
        from .src import train
        sys.exit(train.main(arg_list[2:]))
    if sys.argv[1] == 'infer':
        from .src import inference
        sys.exit(inference.main(arg_list[2:]))
    log.error(f'Invalid argument {arg_list[1]}. Please run either "neural-admixture train" or "neural-admixture infer"')
    sys.exit(1)



import numpy as np
import sys

TOL = 1e-4

def run_checks():
    # Load actual outputs
    try:
        Q = np.genfromtxt('./outputs/demo_run.7.Q')
        P = np.genfromtxt('./outputs/demo_run.7.P')
    except FileNotFoundError as e:
        print('Could not find output files. Please make sure to run the demo before running the test.')
        return False
    except Exception as e:
        raise e
    # Load expected outputs
    try:
        expected_Q = np.genfromtxt('./outputs/demo_run.7.Q.expected')
        expected_P = np.genfromtxt('./outputs/demo_run.7.P.expected')
    except FileNotFoundError as e:
        print('Could not expected output files. Please make sure they are present in the outputs folder of the demo.')
        return False
    except Exception as e:
        raise e
    # Check mean error is below tolerance
    return np.allclose(Q, expected_Q) and np.allclose(P, expected_P)

if __name__ == '__main__':
    passed = run_checks()
    print(f'Output and expected output are {"" if passed else "NOT "}similar.')
    sys.exit(0)



import json
import logging
import sys
import torch
from typing import List

from . import utils
from ..model.neural_admixture import NeuralAdmixture

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

def main(argv: List[str]):
    """Inference entry point
    """
    args = utils.parse_infer_args(argv)
    log.info('Will use GPU' if torch.cuda.is_available() else 'No GPUs available.')
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    data_file_str = args.data_path
    out_name = args.out_name
    weights_file_str = f'{args.save_dir}/{args.name}.pt'
    config_file_str = f'{args.save_dir}/{args.name}_config.json'
    try:
        with open(config_file_str, 'r') as fb:
            config = json.load(fb)
    except FileNotFoundError as fnfe:
        log.error(f'Config file ({config_file_str}) not found. Make sure it is in the correct directory and with the correct name.')
        return 1
    except Exception as e:
        raise e
    log.info('Model config file loaded. Loading weights...')
    model = NeuralAdmixture(config['Ks'], num_features=config['num_snps'])
    model.load_state_dict(torch.load(weights_file_str, map_location=device), strict=True)
    model.to(device)
    log.info('Model weights loaded.')
    X, _, _, _ = utils.read_data(data_file_str)
    assert X.shape[1] == config['num_snps'], 'Number of SNPs in data does not correspond to number of SNPs the network was trained on.'
    log.info('Data loaded and validated. Running inference...')
    _ = utils.get_model_predictions(model, X, bsize=args.batch_size, device=device)
    log.info('Inference run successfully. Writing outputs...')
    utils.write_outputs(model, X, valX=None, bsize=args.batch_size,
                        device=device, run_name=out_name,
                        out_path=args.save_dir, only_Q=True)
    log.info('Exiting...')
    logging.shutdown()
    return 0



import torch.nn as nn
import torch.optim as optim

from . import initializations as init

class Switchers(object):
    """Switcher object for several utilities
    """
    _activations = {
        'relu': lambda x: nn.ReLU(),
        'tanh': lambda x: nn.Tanh(),
        'gelu': lambda x: nn.GELU()
    }

    _initializations = {
        'pckmeans': lambda X, y, k, seed, path, run_name, n_comp, batch_size: init.PCKMeansInitialization.get_decoder_init(X, k, path, run_name, n_comp, seed, batch_size),
        'pcarchetypal': lambda X, y, k, seed, path, run_name, n_comp, batch_size: init.PCArchetypal.get_decoder_init(X, k, path, run_name, n_comp, seed, batch_size),
        'pretrained': lambda X, y, k, seed, path, run_name, n_comp, batch_size: init.PretrainedInitialization.get_decoder_init(X, k, path),
        'supervised': lambda X, y, k, seed, path, run_name, n_comp, batch_size: init.SupervisedInitialization.get_decoder_init(X, y, k)
    }

    _optimizers = {
        'adam': lambda params, lr: optim.Adam(params, lr),
        'sgd': lambda params, lr: optim.SGD(params, lr),
    }

    @classmethod
    def get_switchers(cls):
        return {
            'activations': cls._activations,
            'initializations': cls._initializations,
            'optimizers': cls._optimizers
        }



import logging
import sys
import torch
import torch.nn as nn

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

class ZeroOneClipper(object):
    def __call__(self, module):
        if hasattr(module, 'weight'):
            with torch.no_grad():
                w = module.weight.data
                w.clamp_(0.,1.) # 0 <= F_ij <= 1

class NeuralEncoder(nn.Module):
    def __init__(self, input_size, ks):
        super().__init__()
        assert sum([k < 2 for k in ks]) == 0, 'Invalid number of clusters. Requirement: k >= 2'
        self.ks = ks
        self.heads = nn.ModuleList([
            nn.Sequential(
                nn.Linear(input_size, k, bias=True)
                # nn.BatchNorm1d(k)
            ) for k in ks])

    def _get_head_for_k(self, k):
        return self.heads[k-min(self.ks)]

    def forward(self, X):
        outputs = [self._get_head_for_k(k)(X) for k in self.ks]
        return outputs


class NeuralDecoder(nn.Module):
    def __init__(self, ks, output_size, bias=False, inits=None, freeze=False):
        super().__init__()
        self.ks = ks
        self.force_freeze = freeze
        if inits is None:
            self.decoders = nn.ModuleList(
                [nn.Linear(k, output_size, bias=bias) for k in self.ks]
            )
            if freeze:
                log.warn('Not going to freeze weights as no initialization was provided.')   
        else:
            layers = [None]*len(self.ks)
            for i in range(len(ks)):
                ini = end if i != 0 else 0
                end = ini+self.ks[i]
                layers[i] = nn.Linear(self.ks[i], output_size, bias=bias)
                layers[i].weight = torch.nn.Parameter(inits[ini:end].clone().detach().T)
            self.decoders = nn.ModuleList(layers)
            if self.force_freeze:
                log.info('Decoder weights will be frozen.')
                self.freeze()
        assert len(self.decoders) == len(self.ks)

    def _get_decoder_for_k(self, k):
        return self.decoders[k-min(self.ks)]

    def forward(self, hid_states):
        outputs = [torch.clamp(self._get_decoder_for_k(self.ks[i])(hid_states[i]), 0, 1) for i in range(len(self.ks))]
        return outputs
    
    def freeze(self) -> None:
        for param in self.parameters():
            param.requires_grad = False
        return
    
    def unfreeze(self) -> None:
        for param in self.parameters():
            param.requires_grad = True
        return



import pytest

from neural_admixture.model.switchers import Switchers


def test_initialization_switcher_nargs(mocker):
    mocker.patch("neural_admixture.model.initializations.PCKMeansInitialization.get_decoder_init", return_value=0)
    mocker.patch("neural_admixture.model.initializations.PCArchetypal.get_decoder_init", return_value=0)
    mocker.patch("neural_admixture.model.initializations.SupervisedInitialization.get_decoder_init", return_value=0)
    mocker.patch("neural_admixture.model.initializations.PretrainedInitialization.get_decoder_init", return_value=0)
    n_switcher_args = [None]*8
    assert Switchers().get_switchers()["initializations"]["pckmeans"](*n_switcher_args) == 0
    assert Switchers().get_switchers()["initializations"]["pcarchetypal"](*n_switcher_args) == 0
    assert Switchers().get_switchers()["initializations"]["supervised"](*n_switcher_args) == 0
    assert Switchers().get_switchers()["initializations"]["pretrained"](*n_switcher_args) == 0


import argparse
import configargparse
import logging
import numpy as np
import os
import sys
import torch
import wandb

import dask.array as da
from pathlib import Path
from typing import Dict, List, Tuple, Union

from .snp_reader import SNPReader
from ..model.neural_admixture import NeuralAdmixture

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

def parse_train_args(argv: List[str]):
    """Training arguments parser
    """
    parser = configargparse.ArgumentParser(prog='neural-admixture train',
                                           description='Rapid population clustering with autoencoders - training mode',
                                           config_file_parser_class=configargparse.YAMLConfigFileParser)
    parser.add_argument('--learning_rate', required=False, default=1e-5, type=float, help='Learning rate')
    parser.add_argument('--max_epochs', required=False, type=int, default=50, help='Maximum number of epochs')
    parser.add_argument('--initialization', required=False, type=str, default = 'pckmeans',
                        choices=['pretrained', 'pckmeans', 'supervised', 'pcarchetypal'],
                        help='Decoder initialization (overriden if supervised)')
    parser.add_argument('--optimizer', required=False, default='adam', type=str, choices=['adam', 'sgd'], help='Optimizer')
    parser.add_argument('--save_every', required=False, default=10, type=int, help='Save checkpoint every this number of epochs')
    parser.add_argument('--l2_penalty', required=False, default=0.0005, type=float, help='L2 penalty on encoder weights')
    parser.add_argument('--activation', required=False, default='gelu', type=str, choices=['relu', 'tanh', 'gelu'], help='Activation function for encoder layers')
    parser.add_argument('--seed', required=False, type=int, default=42, help='Seed')
    parser.add_argument('--k', required=False, type=int, help='Number of populations/clusters (single-head)')
    parser.add_argument('--min_k', required=False, type=int, help='Minimum number of populations/clusters (multi-head)')
    parser.add_argument('--max_k', required=False, type=int, help='Maximum number of populations/clusters (multi-head)')
    parser.add_argument('--hidden_size', required=False, default=64, type=int, help='Dimension of first projection in encoder')
    parser.add_argument('--init_file', required=False, type=str, help='File name of precomputed initialization weights to load from/save to')
    parser.add_argument('--freeze_decoder', action='store_true', default=False, help='If specified, will freeze decoder weights during training (only use if providing a computed initialization!)')
    parser.add_argument('--supervised', action='store_true', default=False, help='If specified, will run the supervised version')
    parser.add_argument('--validation_data_path', required=False, default='', type=str, help='Path containing the validation data')
    parser.add_argument('--populations_path', required=False, default='', type=str, help='Path containing the main data populations')
    parser.add_argument('--validation_populations_path', required=False, default='', type=str, help='Path containing the validation data populations')
    parser.add_argument('--wandb_log', required=False, action='store_true', default=False, help='Whether to log to wandb or not')
    parser.add_argument('--wandb_user', required=False, type=str, help='wandb user')
    parser.add_argument('--wandb_project', required=False, type=str, help='wandb project')
    parser.add_argument('--pca_path', required=False, type=str, help='Path containing PCA object, used for plots and to store checkpoints')
    parser.add_argument('--pca_components', required=False, type=int, default=2, help='Number of components to use for the PCKMeans initialization')
    parser.add_argument('--tol', required=False, type=float, default=1e-6, help='Convergence criterion: will stop when difference in objective function between two iterations is smaller than this')
    parser.add_argument('--save_dir', required=True, type=str, help='Save model in this directory')
    parser.add_argument('--data_path', required=True, type=str, help='Path containing the main data')
    parser.add_argument('--name', required=True, type=str, help='Experiment/model name')
    parser.add_argument('--batch_size', required=False, default=400, type=int, help='Batch size')
    parser.add_argument('--supervised_loss_weight', required=False, default=0.05, type=float, help='Weight given to the supervised loss')
    parser.add_argument('--warmup_epochs', required=False, default=10, type=int, help='Number of warmup epochs to bring Q to a good initialization. Set to 0 to skip warmup.')
    # parser.add_argument('--cv', required=False, default=None, type=int, help='Number of folds for cross-validation')
    return parser.parse_args(argv)

def parse_infer_args(argv: List[str]):
    """Inference arguments parser
    """
    parser = argparse.ArgumentParser(prog='neural-admixture infer',
                                     description='Rapid population clustering with autoencoders - inference mode')
    parser.add_argument('--out_name', required=True, type=str, help='Name used to output files on inference mode')
    parser.add_argument('--save_dir', required=True, type=str, help='Load model from this directory')
    parser.add_argument('--data_path', required=True, type=str, help='Path containing the main data')
    parser.add_argument('--name', required=True, type=str, help='Trained experiment/model name')
    parser.add_argument('--batch_size', required=False, default=400, type=int, help='Batch size')
    return parser.parse_args(argv)

def initialize_wandb(run_name: str, trX: da.core.Array, valX: da.core.Array, args: argparse.Namespace, out_path: str, silent: bool=True) -> str:
    """Initializes wandb project run

    Args:
        run_name (str): run name.
        trX (da.core.Array): Dask array containing training data.
        valX (da.core.Array): Dask array containing validation data.
        args (argparse.Namespace): parsed arguments.
        out_path (str): results output path.
        silent (bool, optional): if set, skips wandb log messages. Defaults to True.

    Returns:
        str: run name.
    """
    if run_name is None:
        log.warn('Run name for wandb not specified. Skipping.')
        return None
    if silent:
        os.environ['WANDB_SILENT'] = 'true'
    wandb.init(project=args.wandb_project,
               entity=args.wandb_user,
               name=run_name,
               config=args,
               settings=wandb.Settings(start_method='fork')
            )
    wandb.config.update({'train_samples': trX.shape[0],
                         'val_samples': valX.shape[0] if valX is not None else '',
                         'SNPs': trX.shape[1],
                         'out_path': out_path})
    return run_name

def read_data(tr_file: str, val_file: str=None, tr_pops_f: str=None, val_pops_f: str=None) -> Tuple[da.core.Array, Union[None, da.core.Array], Union[None, List[str]], Union[None, List[str]]]:
    """Read data in any compatible format

    Args:
        tr_file (str): denotes the path of the main data file
        val_file (str, optional): denotes the path of the validation data file. Defaults to None.
        tr_pops_f (str, optional): denotes the path containing the main populations file. Defaults to None.
        val_pops_f (str, optional): denotes the path containing the validation populations file. Defaults to None.

    Returns:
        Tuple[da.core.Array, Union[None, da.core.Array], Union[None, List[str]], Union[None, List[str]]]: data objects
    """
    ''' 
    tr_file: string denoting the path of the main data file. The format of this file must be either HDF5 or VCF.
    val_file: optional. String denoting the path of the validation data file. The format of this file must be either HDF5 or VCF.
    tr_pops_f: optional. String denoting the path containing the main populations file. It must be a plain txt file where each row is a number specifying the population of the corresponding sample, as in ADMIXTURE.
    val_pops_f: optional. String denoting the path containing the validation populations file. It must be a plain txt file where each row is a number specifying the population of the corresponding sample, as in ADMIXTURE.
    '''
    tr_pops, val_pops = None, None
    log.info('Reading data...')
    snp_reader = SNPReader()
    tr_snps = snp_reader.read_data(tr_file)
    val_snps = snp_reader.read_data(val_file) if val_file else None
    if tr_pops_f:
        with open(tr_pops_f, 'r') as fb:
            tr_pops = [p.strip() for p in fb.readlines()]
    if val_pops_f:
        with open(val_pops_f, 'r') as fb:
            val_pops = [p.strip() for p in fb.readlines()]
    validate_data(tr_snps, tr_pops, val_snps, val_pops)
    log.info(f'Data contains {tr_snps.shape[0]} samples and {tr_snps.shape[1]} SNPs.')
    if val_snps is not None:
        log.info(f'Validation data contains {val_snps.shape[0]} samples and {val_snps.shape[1]} SNPs.')
    log.info('Data loaded.')
    return tr_snps, tr_pops, val_snps, val_pops

def validate_data(tr_snps: da.core.Array, tr_pops: Union[None, List[str]], val_snps: Union[None, da.core.Array], val_pops: Union[None, List[str]]):
    """Data sanity check

    Args:
        tr_snps (da.core.Array): Dask array containing training data.
        tr_pops (Union[None, List[str]]): list containing training population labels.
        val_snps (Union[None, da.core.Array]): Dask array containing validation data.
        val_pops (Union[None, List[str]]): list containing validation population labels.
    """
    assert not (val_snps is None and val_pops is not None), 'Populations were specified for validation data, but no SNPs were specified.'
    if tr_pops is not None:
        assert tr_snps.shape[0] == len(tr_pops), f'Number of samples in data and population file does not match: {len(tr_snps)} vs {len(tr_pops)}.'
    if val_snps is not None:
        assert tr_snps.shape[1] == val_snps.shape[1], f'Number of SNPs in training and validation data does not match: {tr_snps.shape[1]} vs {val_snps.shape[1]}.'
        if val_pops is not None:
            assert val_snps.shape[0] == len(val_pops), f'Number of samples in validation data and validation population file does not match: {len(tr_snps)} vs {len(tr_pops)}.'
    return

def get_model_predictions(model: NeuralAdmixture, data: da.core.Array, bsize: int, device: torch.device) -> List[np.ndarray]:
    """Helper function to run inference on data

    Args:
        model (NeuralAdmixture): trained model object.
        data (da.core.Array): Dask array containing data to get results from.
        bsize (int): batch size.
        device (torch.device): torch device.

    Returns:
        List[np.ndarray]: list of Q matrices
    """
    model.to(torch.device(device))
    outs = [torch.tensor([]) for _ in range(len(model.ks))]
    model.eval()
    with torch.inference_mode():
        for X, _, _ in model.batch_generator(data, bsize, shuffle=False):
            X = X.to(device)
            out = model(X, True)
            for j in range(len(model.ks)):
                outs[j] = torch.cat((outs[j], out[j].detach().cpu()), axis=0)
    return [out.cpu().numpy() for out in outs]

def get_model_reconstructions(model: NeuralAdmixture, data: da.core.Array, bsize: int, device: torch.device) -> List[np.ndarray]:
    """Helper function to run inference on data

    Args:
        model (NeuralAdmixture): trained model object.
        data (da.core.Array): Dask array containing data to get results from.
        bsize (int): batch size.
        device (torch.device): torch device.

    Returns:
        List[np.ndarray]: list of reconstructions (matrix product of Q and P) for different Ks
    """
    model.to(torch.device(device))
    outs = [torch.tensor([]) for _ in range(len(model.ks))]
    model.eval()
    with torch.inference_mode():
        for X, _, _ in model.batch_generator(data, bsize, shuffle=False):
            X = X.to(device)
            out, _ = model(X)
            for j in range(len(model.ks)):
                outs[j] = torch.cat((outs[j], out[j].detach().cpu()), axis=0)
    return [out.cpu().numpy() for out in outs]

def write_outputs(model: NeuralAdmixture, trX: da.core.Array, valX: Union[da.core.Array, None],
                  bsize: int, device: torch.device, run_name: str, out_path: str, only_Q: bool=False) -> int:
    """Helper function to write Q and P matrices to disk

    Args:
        model (NeuralAdmixture): trained model object.
        trX (da.core.Array): Dask array containing training data.
        valX (Union[da.core.Array, None]): Dask array containing validation data.
        bsize (int): batch size.
        device (torch.device): torch device.
        run_name (str): run name.
        out_path (str): output directory path.
        only_Q (bool, optional): if set, only the Q matrix will be written. Defaults to False.

    Returns:
        int: status
    """
    out_path = Path(out_path)
    if not only_Q:
        for dec in model.decoders.decoders:
            w = 1-dec.weight.data.cpu().numpy()
            k = dec.in_features
            np.savetxt(out_path/f"{run_name}.{k}.P", w, delimiter=' ')
    tr_preds = get_model_predictions(model, trX, bsize, device)
    for i, k in enumerate(model.ks):
        np.savetxt(out_path/f"{run_name}.{k}.Q", tr_preds[i], delimiter=' ')
    if valX is not None:
        val_preds = get_model_predictions(model, valX, bsize, device)
        for i, k in enumerate(model.ks):
            np.savetxt(out_path/f"{run_name}_validation.{k}.Q", val_preds[i], delimiter=' ')
    return 0

def compute_deviances(model: NeuralAdmixture, data: da.core.Array, bsize: int, device: torch.device) -> Dict[int, float]:
    """_summary_

    Args:
        model (NeuralAdmixture): _description_
        trX (np.ndarray): _description_
        bsize (int): _description_
        device (torch.device): _description_

    Returns:
        Dict[int, float]: _description_
    """
    eps = 1e-7
    reconstructions = get_model_reconstructions(model, data, bsize, device)
    data_tmp = 2*data # Multiply by 2 to get to same reconstruction range to ADMIXTURE
    data_tmp = da.clip(data_tmp, eps, 2-eps) # To avoid log(0)
    deviances = {}
    for rec, K in zip(reconstructions, model.ks):
        rec *= 2 # Multiply by 2 to get to same reconstruction range to ADMIXTURE
        rec = da.clip(rec, eps, 2-eps) # To avoid division by zero
        deviance = (data_tmp * da.log(data_tmp/rec) + (2-data_tmp) * da.log((2-data_tmp)/(2-rec)))**2
        deviances[K] = da.mean(deviance).compute()
    return deviances



import dask.array as da
import logging
import pytest
import random
import string

from neural_admixture.model.initializations import PCArchetypal, PCKMeansInitialization, SupervisedInitialization


@pytest.mark.parametrize(["n", "d", "K"], [[42, 1000, [2,3,4]], [500, 100, [99,100,101]], [30,100,[2,3,4]]])
def test_pckmeans(n, d, K, caplog):
    caplog.set_level(logging.CRITICAL)
    X = da.random.binomial(n=2, p=.1, size=(n,d))/2
    path = None
    run_name = "pckmeans_test"
    n_components = 4
    seed = 42
    batch_size = 200
    P_init, Q_inits = PCKMeansInitialization().get_decoder_init(X, K, path, run_name, n_components, seed, batch_size)
    assert P_init.shape[0] == sum(K) and P_init.shape[1] == d
    assert len(Q_inits) == len(K)
    assert all([Q_init.shape[0] == n and Q_init.shape[1] == k for Q_init, k in zip(Q_inits, K)])

@pytest.mark.parametrize(["n", "d", "K"], [[42, 1000, [2,3,4]], [500, 100, [3,4,5]], [30,100,[6,7,8]]])
def test_pcarchetypal(n, d, K, caplog):
    caplog.set_level(logging.CRITICAL)
    X = da.random.binomial(n=2, p=.1, size=(n,d))/2
    path = None
    run_name = "test_pcarchetypal"
    n_components = 4
    seed = 42
    batch_size = 200
    P_init, _ = PCArchetypal().get_decoder_init(X, K, path, run_name, n_components, seed, batch_size)
    assert P_init.shape[0] == sum(K) and P_init.shape[1] == d

@pytest.mark.parametrize(["n", "d", "K"], [[42, 1000, [2]], [500, 100, [5]], [30,100,[10]]])
def test_supervised(n, d, K, caplog):
    caplog.set_level(logging.CRITICAL)
    X = da.random.binomial(n=2, p=.1, size=(n,d))/2
    ancs = random.sample(string.ascii_lowercase, K[0]) + ["-"]
    y = ancs+random.choices(ancs, k=n-len(ancs))
    P_init, _ = SupervisedInitialization().get_decoder_init(X, y, K)
    assert P_init.shape[0] == K[0] and P_init.shape[1] == d

def test_supervised_fail(caplog):
    caplog.set_level(logging.CRITICAL)
    n, d = 100, 10
    K = [5,6,7]
    X = da.random.binomial(n=2, p=.1, size=(n,d))/2
    ancs = random.sample(string.ascii_lowercase, 5) + ["-"]
    y = random.choices(ancs, k=n)
    with pytest.raises(NotImplementedError):
        _, _ = SupervisedInitialization().get_decoder_init(X, y, K)



import dask
import dask.array as da
import json
import logging
import math
import numpy as np
import random
import sys
import torch
import torch.nn as nn
import wandb
from pathlib import Path
from tqdm.auto import tqdm
from typing import Iterable, Optional, Tuple, Union

from .modules import NeuralDecoder, NeuralEncoder, ZeroOneClipper

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

class NeuralAdmixture(nn.Module):
    """Instantiate Neural ADMIXTURE model

    Args:
        ks (Iterable[int]): different numbers of populations to use. If the list contains more than one value, the multi-head version will run.
        num_features (int): number of SNPs used to train the network.
        encoder_activation (nn.Module, optional): activation function used in the encoder. Defaults to nn.GELU().
        P_init (Optional[torch.Tensor], optional): if provided, corresponds to initialization weights as returned by one of the initialization functions. Defaults to None.
        lambda_l2 (float, optional): L2 regularization strength in the encoder. Smaller values will give quasi-binary predictions. Defaults to 5e-4.
        hidden_size (int, optional): number of neurons in the first linear layer. Defaults to 64.
        freeze_decoder (bool, optional): if set to True, the decoder weights are frozen (useful for reusing other results). Defaults to False.
        supervised (bool, optional): if set to True, will run in supervised mode. Defaults to False.
        supervised_loss_weight (float, optional): weight given to the supervised loss term. Only applied if running in supervised mode. Defaults to 0.05.
    """
    def __init__(self, ks: Iterable[int], num_features: int, encoder_activation: nn.Module=nn.GELU(),
                 P_init: Optional[torch.Tensor]=None, lambda_l2: float=5e-4, hidden_size: int=64,
                 freeze_decoder: bool=False, supervised: bool=False, supervised_loss_weight: float=0.05) -> None:
        super().__init__()
        self.ks = ks
        self.num_features = num_features
        self.hidden_size = hidden_size
        self.encoder_activation = encoder_activation
        self.supervised = supervised
        self.supervised_loss_weight = supervised_loss_weight
        self.freeze_decoder = freeze_decoder
        self.batch_norm = nn.BatchNorm1d(self.num_features)
        self.lambda_l2 = lambda_l2 if lambda_l2 > 1e-8 else 0
        self.softmax = nn.Softmax(dim=1)
        self.common_encoder = nn.Sequential(
                nn.Linear(self.num_features, self.hidden_size, bias=True),
                self.encoder_activation,
        )
        self.multihead_encoder = NeuralEncoder(self.hidden_size, self.ks)
        self.decoders = NeuralDecoder(self.ks, num_features, bias=False, inits=P_init, freeze=self.freeze_decoder)
        self.clipper = ZeroOneClipper()
        self.decoders.decoders.apply(self.clipper)

    def forward(self, X: torch.Tensor, only_assignments: bool=False,
                only_hidden_states: bool=False) -> Union[torch.Tensor, Iterable[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        """Forward pass of the model

        Args:
            X (torch.Tensor): _description_
            only_assignments (bool, optional): if set to True, only return the Q values (used in inference mode for speeding up). Defaults to False.
            only_hidden_states (bool, optional): if set to True, only return logits of the Q values. Defaults to False.

        Returns:
            Union[torch.Tensor, Iterable[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]: SNP reconstruction and Q values (if last two parameters are set to False), otherwise only Q values or logits of Q values.
        """
        X = self.batch_norm(X)
        enc = self.common_encoder(X)
        del X
        hid_states = self.multihead_encoder(enc)
        if only_hidden_states:
            return hid_states
        probs = [self.softmax(h) for h in hid_states]
        if only_assignments:
            return probs
        del enc
        return self.decoders(probs), probs

    def launch_training(self, trX: da.core.Array, optimizer: torch.optim.Optimizer,
                        loss_f: torch.nn.modules.loss._Loss, num_epochs: int,
                        device: torch.device, batch_size: int=0, valX: Optional[da.core.Array]=None,
                        save_every: int=10, save_path: str='../outputs/model.pt',
                        trY: Optional[Iterable[str]]=None, valY: Optional[Iterable[str]]=None, seed: int=42,
                        shuffle: bool=True, log_to_wandb: bool=False, tol: float=1e-5, dry_run: bool=False,
                        warmup_epochs: int=10, Q_inits: Optional[Iterable[torch.Tensor]]=None) -> int:
        """Launch training pipeline of the model

        Args:
            trX (da.core.Array): training data matrix
            optimizer (torch.optim.Optimizer): loaded optimizer to use
            loss_f (torch.nn.modules.loss._Loss): instantiated loss function to use
            num_epochs (int): maximum number of epochs. Note that the training might stop earlier if the loss is in a plateau.
            device (torch.device): device to use for training
            batch_size (int, optional): batch size. If 0, will use all the training data in a single batch. Defaults to 0.
            valX (Optional[da.core.Array], optional): _description_. Defaults to None.
            save_every (int, optional): save a checkpoint after this number of epochs. Defaults to 10.
            save_path (str, optional): output path of the trained model. Defaults to '../outputs/model.pt'.
            trY (Optional[Iterable[str]], optional): list of training populations per sample. Defaults to None.
            valY (Optional[Iterable[str]], optional): list of validation populations per sample. Defaults to None.
            seed (int, optional): seed for RNG. Defaults to 42.
            shuffle (bool, optional): whether to shuffle the samples at every epoch. Defaults to True.
            log_to_wandb (bool, optional): whether to log training to wandb. Defaults to False.
            tol (float, optional): tolerance for early stopping. Training will stop if decrease in loss is smaller than this value. Defaults to 1e-5.
            dry_run (bool, optional): whether to run a dry run (no output is written to disk). Defaults to False.
            warmup_epochs (int, optional): number of warmup epochs to bring Q to a good initial solution. If 0, no warmup is performed. Defaults to 10.
            Q_inits(Optional[Iterable[torch.Tensor]], optional): initial Q values for warmup. Defaults to None.
        Returns:
            int: number of actual training epochs ran
        """
        random.seed(seed)
        loss_f_supervised, trY_num, valY_num = None, None, None
        tr_losses, val_losses = [], []
        log.info(f'Will stop optimization when difference in objective function between two subsequent iterations is < {tol} or after {num_epochs} epochs.')
        if self.supervised:
            log.info('Going to train on supervised mode.')
            assert trY is not None, 'Training ground truth ancestries needed for supervised mode'
            ancestry_dict = {anc: idx for idx, anc in enumerate(sorted(np.unique([a for a in trY if a  != '-'])))}
            assert len(ancestry_dict) == self.ks[0], 'Number of ancestries in training ground truth is not equal to the value of k'
            ancestry_dict['-'] = -1
            to_idx_mapper = np.vectorize(lambda x: ancestry_dict[x])
            trY_num = to_idx_mapper(trY[:])
            valY_num = to_idx_mapper(valY[:]) if valY is not None else None
            loss_f_supervised = nn.CrossEntropyLoss(reduction='mean')
        log.info("Bringing training data into memory...")
        trX = trX.compute()
        log.info("Running warmup epochs...")
        if warmup_epochs > 0:
            if Q_inits is None:
                log.warning("No Q initialization was provided. Skipping warmup epochs.")
            else:
                self.decoders.freeze()
                loss_f_warmup = nn.BCELoss(reduction='mean')
                opt_warmup = torch.optim.AdamW(self.common_encoder.parameters(), lr=1e-5)
                for wep in range(warmup_epochs):
                    _, _ = self._run_warmup_epoch(trX, Q_inits, opt_warmup, loss_f_warmup, batch_size, device, shuffle, epoch_num=wep+1)
                if not self.decoders.force_freeze:
                    self.decoders.unfreeze()
        log.info("Training...")
        for ep in range(num_epochs):
            tr_loss, val_loss = self._run_epoch(trX, optimizer, loss_f, batch_size, valX, device, shuffle, loss_f_supervised, trY_num, valY_num, epoch_num=ep+1)
            tr_losses.append(tr_loss)
            val_losses.append(val_loss)
            assert not math.isnan(tr_loss), 'Training loss is NaN'
            if log_to_wandb and val_loss is not None:
                wandb.log({"tr_loss": tr_loss, "val_loss": val_loss})
            elif log_to_wandb and val_loss is None:
                wandb.log({"tr_loss": tr_loss})
            tr_diff = tr_losses[-2]-tr_losses[-1] if len(tr_losses) > 1 else 'NaN'
            val_diff = val_losses[-2]-val_losses[-1] if val_loss is not None and len(val_losses) > 1 else 'NaN'
            log.info(f'[METRICS] EPOCH {ep+1}: mean training loss: {tr_loss}, diff: {tr_diff}')
            if val_loss is not None:
                log.info(f'[METRICS] EPOCH {ep+1}: mean validation loss: {val_loss}, diff: {val_diff}')
            if not dry_run and save_every*ep > 0 and ep % save_every == 0:
                torch.save(self.state_dict(), save_path)
            if ep > 15 and tol > 0 and self._has_converged(tr_diff, tol):
                log.info(f'Convergence criteria met. Stopping fit after {ep+1} epochs...')
                return ep+1
        log.info(f'Max epochs reached. Stopping fit...')
        return ep+1

    def _get_encoder_norm(self, p: int=2) -> torch.Tensor:
        """Retrieve the sum of the norm of the encoder parameters (for regularization purposes)

        Args:
            p (int, optional): norm to retrieve. Defaults to 2.

        Returns:
            torch.Tensor: sum of norms of the encoder parameters
        """
        shared_params = torch.cat([x.view(-1) for x in self.common_encoder.parameters()])
        multihead_params = torch.cat([x.view(-1) for x in self.multihead_encoder.parameters()])
        return torch.norm(shared_params, p)+torch.norm(multihead_params, p)

    def _run_step(self, X: torch.Tensor, optimizer: torch.optim.Optimizer, loss_f: torch.nn.modules.loss._Loss,
                  loss_f_supervised: Optional[torch.nn.modules.loss._Loss], y: Optional[Iterable[str]],
                  warmup: Optional[bool]=False, Q_inits: Optional[Iterable[torch.Tensor]]=None) -> float:
        """Run a single optimization step
        Args:
            X (torch.Tensor): mini-batch of data
            optimizer (torch.optim.Optimizer): loaded optimizer to use
            loss_f (torch.nn.modules.loss._Loss): instantiated loss function to use
            loss_f_supervised (Optional[torch.nn.modules.loss._Loss]): instantiated supervied loss function to use
            y (Optional[Iterable[str]], optional): list of training populations per sample. Defaults to None.
            warmup (Optional[bool], optional): whether to run a warmup step. Defaults to False.
            Q_inits (Optional[Iterable[torch.Tensor]], optional): list of Q initialization matrices. Defaults to None.

        Returns:
            float: loss value of the mini-batch
        """
        optimizer.zero_grad(set_to_none=True)
        if warmup:
            hid_states = self(X, only_assignments=True)
            recs = None
        else:
            recs, hid_states = self(X)
        if Q_inits is None and not warmup: # Regular step
            loss = sum((loss_f(rec, X) for rec in recs))
        elif Q_inits is not None and warmup: # Warmup step
            loss = sum((loss_f(h, Q_init) for h, Q_init in zip(hid_states, Q_inits)))
        else:
            raise ValueError("Cannot provide Q initialization for a regular step")
        if loss_f_supervised is not None:  # Currently only implemented for single-head architecture!
            mask = y > -1
            loss += sum((self.supervised_loss_weight*loss_f_supervised(h[mask], y[mask]) for h in hid_states))
        if not warmup and self.lambda_l2 > 0:
            loss += self.lambda_l2*self._get_encoder_norm(2)**2
        del recs, hid_states
        loss.backward()
        optimizer.step()
        self.decoders.decoders.apply(self.clipper)
        return loss.item()
    
    def _validate(self, valX: da.core.Array, loss_f: torch.nn.modules.loss._Loss, batch_size: int, device: torch.device,
                  loss_f_supervised: Optional[torch.nn.modules.loss._Loss]=None, y: Optional[Iterable[str]]=None) -> float:
        """_summary_

        Args:
            valX (da.core.Array): validation data matrix
            loss_f (torch.nn.modules.loss._Loss): instantiated loss function to use
            batch_size (int): batch size. Note that the last batch might be smaller.
            device (torch.device): device to use for training
            loss_f_supervied (Optional[torch.nn.modules.loss._Loss]): instantiated supervied loss function to use
            y (Optional[Iterable[str]], optional): list of training populations per sample. Defaults to None.

        Returns:
            float: loss value of the validation set
        """
        acum_val_loss = 0
        with torch.inference_mode():
            for X, y_b in self.batch_generator(valX, batch_size, y=y if loss_f_supervised is not None else None):
                X = X.to(device)
                y_b = y_b.to(device) if y_b is not None else None
                recs, hid_states = self(X)
                acum_val_loss += sum((loss_f(rec, X).item() for rec in recs))
                if loss_f_supervised is not None and y_b is not None:
                    mask = y_b > -1
                    acum_val_loss += sum((self.supervised_loss_weight*loss_f_supervised(h[mask], y_b[mask]).item() for h in hid_states))
            if self.lambda_l2 > 1e-6:
                acum_val_loss += self.lambda_l2*self._get_encoder_norm()**2
        return acum_val_loss

    def batch_generator(self, X, batch_size=0, shuffle=True, y=None, Q_inits=None):
        is_inmem = not isinstance(X, da.core.Array)
        idxs = [i for i in range(X.shape[0])]
        if shuffle:
            random.shuffle(idxs)
        if batch_size < 1:
            batch_size = X.shape[0]
        else:
            for i in range(0, X.shape[0], batch_size):
                with dask.config.set(**{'array.slicing.split_large_chunks': True}):
                    to_yield_X, to_yield_y, to_yield_Y_warmup = X[sorted(idxs[i:i+batch_size])], None, None
                if not is_inmem:
                    to_yield_X = to_yield_X.compute()
                if y is not None:
                    to_yield_y = torch.as_tensor(y[sorted(idxs[i:i+batch_size])], dtype=torch.int64)
                if Q_inits is not None:
                    to_yield_Y_warmup = [Q[sorted(idxs[i:i+batch_size])] for Q in Q_inits]
                yield torch.as_tensor(to_yield_X, dtype=torch.float32), to_yield_y, to_yield_Y_warmup

    def _run_epoch(self, trX, optimizer, loss_f, batch_size, valX,
                   device, shuffle=True, loss_f_supervised=None,
                   trY=None, valY=None, epoch_num=0):
        tr_loss, val_loss = 0, None
        self.train()
        total_b = trX.shape[0]//batch_size+1
        for X, y, _ in tqdm(self.batch_generator(trX, batch_size, shuffle=shuffle, y=trY if loss_f_supervised is not None else None), desc=f"Epoch {epoch_num}", total=total_b):
            step_loss = self._run_step(X.to(device), optimizer, loss_f, loss_f_supervised, y.to(device) if y is not None else None)
            tr_loss += step_loss
        if valX is not None:
            self.eval()
            val_loss = self._validate(valX, loss_f, batch_size, device, loss_f_supervised, valY)
            return tr_loss/trX.shape[0], val_loss/valX.shape[0]
        return tr_loss/trX.shape[0], None

    def _run_warmup_epoch(self, trX, Q_inits, optimizer, loss_f, batch_size,
                          device, shuffle=True, epoch_num=0) -> Tuple[float, Union[float, None]]:
        tr_loss = 0
        total_b = trX.shape[0]//batch_size+1
        self.train()
        for X, _, Ys in tqdm(self.batch_generator(trX, batch_size, shuffle=shuffle, Q_inits=Q_inits), desc=f"Warmup epoch {epoch_num}", total=total_b):
            step_loss = self._run_step(X.to(device), optimizer, loss_f,
                                       None, None, warmup=True,
                                       Q_inits=[Y.to(device) for Y in Ys])
            tr_loss += step_loss
        log.info(f"Warmup training loss: {tr_loss/trX.shape[0]}")# !
        return tr_loss/trX.shape[0], None


    @staticmethod
    def _has_converged(diff, tol):
        return diff < tol
        
    @staticmethod
    def _hudsons_fst(pop1, pop2):
        '''
        Computes Hudson's Fst given variant frequencies of two populations.
        '''
        try:
            num = np.mean(((pop1-pop2)**2))
            den = np.mean(np.multiply(pop1, 1-pop2)+np.multiply(pop2, 1-pop1))+1e-7
            return num/den
        except Exception as e:
            log.error(e)
            return np.nan

    def display_divergences(self):
        for i, k in enumerate(self.ks):
            dec = self.decoders.decoders[i].weight.data.cpu().numpy()
            header = '\t'.join([f'Pop{p}' for p in range(k-1)])
            print(f'\nFst divergences between estimated populations: (K = {k})')
            print(f'\t{header}')
            print('Pop0')
            for j in range(1, k):
                print(f'Pop{j}', end='')
                pop2 = dec[:,j]
                for l in range(j):
                    pop1 = dec[:,l]
                    fst = self._hudsons_fst(pop1, pop2)
                    print(f"\t{fst:0.3f}", end="" if l != j-1 else '\n')
        return
    
    def save_config(self, name, save_dir):
        config = {
            'Ks': self.ks,
            'num_snps': self.num_features,
        }
        with open(Path(save_dir)/f"{name}_config.json", 'w') as fb:
            json.dump(config, fb)
        log.info('Configuration file saved.')
        return



import dask.array as da
import logging
import numpy as np
import sys

from pathlib import Path

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

class SNPReader:
    """Wrapper to read genotype data from several formats
    """
    def _read_vcf(self, file: str) -> np.ndarray:
        """Reader wrapper for VCF files

        Args:
            file (str): path to file.

        Returns:
            np.ndarray: averaged genotype array of shape (n_samples, n_snps)
        """
        log.info('Input format is VCF.')
        import allel
        f_tr = allel.read_vcf(file)
        calldata = f_tr["calldata/GT"]
        if np.isnan(calldata).any():
            log.warning("Data contains missing values. Will perform zero-imputation.")
            calldata = np.nan_to_num(calldata, nan=0.)
        return np.sum(calldata, axis=2).T/2
    
    def _read_hdf5(self, file: str) -> np.ndarray:
        """Reader wrapper for HDF5 files. HDF5 should directly contain the averaged genotype array

        Args:
            file (str): path to file.

        Returns:
            np.ndarray: averaged genotype array of shape (n_samples, n_snps)
        """
        log.info('Input format is HDF5.')
        import h5py
        f_tr = h5py.File(file, 'r')
        return f_tr['snps']
    
    def _read_bed(self, file: str) -> da.core.Array:
        """Reader wrapper for BED files

        Args:
            file (str): path to file.

        Returns:
            da.core.Array: averaged genotype Dask array of shape (n_samples, n_snps)
        """
        log.info('Input format is BED.')
        from pandas_plink import read_plink
        _, _, G = read_plink(str(Path(file).with_suffix("")))
        if da.isnan(G).any().compute():
            log.warning("Data contains missing values. Will perform zero-imputation.")
            G = da.nan_to_num(G, 0.)
        return (G.T/2)
    
    def _read_pgen(self, file: str) -> np.ndarray:
        """Reader wrapper for PGEN files

        Args:
            file (str): path to file

        Returns:
            np.ndarray: averaged genotype array of shape (n_samples, n_snps)
        """
        log.info('Input format is PGEN.')
        try:
            import pgenlib as pg
        except ImportError as ie:
            log.error('Cannot read PGEN file as pgenlib is not installed.')
            sys.exit(1)
        except Exception as e:
            raise e
        pgen = str.encode(file) # Genotype, sample, variant files
        pgen_reader = pg.PgenReader(pgen)
        calldata = np.ascontiguousarray(np.empty((pgen_reader.get_variant_ct(), 2*pgen_reader.get_raw_sample_ct())).astype(np.int32))
        pgen_reader.read_alleles_range(0, pgen_reader.get_variant_ct(), calldata)
        if np.isnan(calldata).any():
            log.warning("Data contains missing values. Will perform zero-imputation.")
            calldata = np.nan_to_num(calldata, nan=0.)
        return (calldata[:,::2]+calldata[:,1::2]).T/2
    
    def _read_npy(self, file: str) -> np.ndarray:
        """Reader wrapper for NPY files

        Args:
            file (str): path to file

        Returns:
            np.ndarray: averaged genotype array of shape (n_samples, n_snps)
        """
        log.info('Input format is NPY.')
        calldata = np.load(file)
        assert calldata.ndim in [2, 3]
        if np.isnan(calldata).any():
            log.warning("Data contains missing values. Will perform zero-imputation.")
            calldata = np.nan_to_num(calldata, nan=0.)
        if calldata.ndim == 2:
            return calldata/2
        return np.nan_to_num(calldata, nan=0.).sum(axis=2)/2
    
    def read_data(self, file: str) -> da.core.Array:
        """Wrapper of readers

        Args:
            file (str): path to file

        Returns:
            da.core.Array: averaged genotype Dask array of shape (n_samples, n_snps)
        """
        file_extensions = Path(file).suffixes
        if '.vcf' in file_extensions:
            G = self._read_vcf(file)
        elif '.h5' in file_extensions or '.hdf5' in file_extensions:
            G = self._read_hdf5(file)
        elif '.bed' in file_extensions:
            G = self._read_bed(file)
        elif '.pgen' in file_extensions:
            G = self._read_pgen(file)
        elif '.npy' in file_extensions:
            G = self._read_npy(file)
        else:
            log.error('Invalid format. Unrecognized file format. Make sure file ends with .vcf | .vcf.gz | .bed | .pgen | .h5 | .hdf5 | .npy')
            sys.exit(1)
        assert int(G.min()) == 0 and int(G.max()) == 1, 'Only biallelic SNPs are supported. Please make sure multiallelic sites have been removed.'
        G_corr = G if np.mean(G) < 0.5 else 1-G
        return G_corr if isinstance(G_corr, da.core.Array) else da.from_array(G_corr)



import dask
import dask.array as da
import logging
import matplotlib.pyplot as plt
import numpy as np
import pickle
import sys
import time
import torch
from collections.abc import Iterable
from dask_ml.decomposition import IncrementalPCA as DaskIncrementalPCA
from sklearn.cluster import KMeans
from typing import Union
from pathlib import Path
from py_pcha import PCHA


logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

def pca_plot(X_pca: np.ndarray, path: str) -> None:
    """Helper function to render a PCA plot

    Args:
        X_pca (np.ndarray): projected data
        path (str): output file path
    """
    plt.figure(figsize=(15,10))
    plt.scatter(X_pca[:,0], X_pca[:,1], s=.9, c='black')
    plt.xticks([])
    plt.yticks([])
    plt.title('Training data projected onto first two components')
    plt.savefig(path)
    log.info('Plot rendered.')
    return

class PCKMeansInitialization(object):
    """PCKMeans initialization
    """
    @classmethod
    def get_decoder_init(cls, X: da.core.Array, K: Union[Iterable, int], path: Union[str, None], run_name: str, n_components: int, seed: int, batch_size: int) -> torch.Tensor:
        """Get decoder initialization weights using PCKMeans

        Args:
            X (da.core.Array): 2D genotype array
            K (Union[Iterable, int]): number of ancestries to run the algorithm on. If an iterable, then assumes multihead version will be run
            path (Union[str, None]): output path
            run_name (str): name characterizing the experiment
            n_components (int): number of components to compute via incremental PCA
            seed (int): RNG seed
            batch_size (int, optional): batch size for incremental PCA. Defaults to 400.

        Returns:
            torch.Tensor: decoder initialization
        """
        log.info('Running PC-KMeans initialization...')
        np.random.seed(seed)
        t0 = time.time()
        try:
            if path is not None:
                with open(path, 'rb') as fb:
                    pca_obj = pickle.load(fb)
                log.info('PCA loaded.')
                if pca_obj.n_components_ != n_components:
                    raise FileNotFoundError
            else:
                raise FileNotFoundError
        except FileNotFoundError as _:
            log.info(f"{n_components}D PCA object not found. Performing IncrementalPCA...")
            pca_obj = DaskIncrementalPCA(n_components=n_components, random_state=42, batch_size=batch_size)
            pca_obj.fit(X)
            if path is not None:
                with open(path, 'wb') as fb:
                    pickle.dump(pca_obj, fb)
        except Exception as e:
            raise e
        assert pca_obj.n_features_ == X.shape[1], 'Computed PCA and training data do not have same number of SNPs' 
        log.info('Projecting data...')
        X_pca = pca_obj.transform(X).compute()
        log.info('Running KMeans on projected data...')
        if isinstance(K, Iterable):
            k_means_objs = [KMeans(n_clusters=i, random_state=42, n_init=10, max_iter=10).fit(X_pca) for i in K]
            centers = np.concatenate([obj.cluster_centers_ for obj in k_means_objs])
            dists = [obj.transform(X_pca) for obj in k_means_objs]
            Q_inits = [torch.as_tensor(-d, dtype=torch.float32).softmax(dim=1) for d in dists]
            P_init = torch.as_tensor(pca_obj.inverse_transform(centers).compute(), dtype=torch.float32).view(sum(K), -1)
        else:
            k_means_obj = KMeans(n_clusters=K, random_state=42, n_init=10, max_iter=10).fit(X_pca)
            dists = k_means_obj.transform(X_pca)
            P_init = torch.as_tensor(pca_obj.inverse_transform(k_means_obj.cluster_centers_).compute(), dtype=torch.float32).view(K, -1)
            Q_inits = [torch.as_tensor(-dists, dtype=torch.float32).softmax(dim=1)]
        te = time.time()
        log.info('Weights initialized in {} seconds.'.format(te-t0))
        log.info('Rendering PCA plot...')
        try:
            if path is not None:
                plot_save_path = Path(path).parent/f"{run_name}_training_pca.png"
                pca_plot(X_pca, plot_save_path)
        except Exception as e:
            log.warn(f'Could not render PCA plot: {e}')
            log.info('Resuming...')
        return P_init, Q_inits

class PCArchetypal(object):
    """PCArchetypal initialization
    """
    @classmethod
    def get_decoder_init(cls, X: da.core.Array, K: Union[Iterable, int], path: str, run_name: str, n_components: int, seed: int, batch_size: int):
        """Get decoder initialization weights using PCArchetypal

        Args:
            X (da.core.Array): 2D genotype array
            K (Union[Iterable, int]): number of ancestries to run the algorithm on. If an iterable, then assumes multihead version will be run
            path (str): output path
            run_name (str): name characterizing the experiment
            n_components (int): number of components to compute via incremental PCA
            seed (int): RNG seed
            batch_size (int, optional): batch size for incremental PCA. Defaults to 400.

        Returns:
            torch.Tensor: decoder initialization
        """
        log.info('Running PCArchetypal initialization...')
        np.random.seed(seed)
        t0 = time.time()
        try:
            if path is not None:
                with open(path, 'rb') as fb:
                    pca_obj = pickle.load(fb)
                log.info('PCA loaded.')
                if pca_obj.n_components_ != n_components:
                    raise FileNotFoundError
            else:
                raise FileNotFoundError
        except FileNotFoundError as fnfe:
            log.info(f'{n_components}D PCA object not found. Performing PCA...')
            pca_obj = DaskIncrementalPCA(n_components=n_components, random_state=42, batch_size=batch_size)
            pca_obj.fit(X)
            if path is not None:
                with open(path, 'wb') as fb:
                    pickle.dump(pca_obj, fb)
        except Exception as e:
            raise e
        assert pca_obj.n_features_ == X.shape[1], 'Computed PCA and training data do not have same number of SNPs'
        log.info(f'Projecting data to {n_components} dimensions...')
        X_proj = pca_obj.transform(X).compute()
        if not isinstance(K, Iterable):
            K = [K]
        log.info(f'Executing archetypal analysis on projected data...')
        archs = np.concatenate([np.array(PCHA(X_proj.T, noc=k, delta=0.001)[0].T) for k in K])
        log.info('Backtransforming archetypes...')
        P_init = torch.as_tensor(pca_obj.inverse_transform(archs).compute(), dtype=torch.float32)
        te = time.time()
        log.info('Weights initialized in {} seconds.'.format(te-t0))
        try:
            if path is not None:
                plot_save_path = Path(path).parent/f"{run_name}_training_pca.png"
                pca_plot(X_proj, plot_save_path)
        except Exception as e:
            log.warn(f'Could not render PCA plot: {e}')
            log.info('Resuming...')
        return P_init, None


class SupervisedInitialization(object):
    """Supervised initialization
    """
    @classmethod
    def get_decoder_init(cls, X, y, K):
        log.info('Running supervised initialization...')
        assert y is not None, 'Ground truth ancestries needed for supervised mode'
        if len(K) > 1:
            raise NotImplementedError("Supervised mode is only available for single-head architectures at the moment.")
        t0 = time.time()
        k = K[0]
        ancestry_dict = {anc: idx for idx, anc in enumerate(sorted(np.unique([a for a in y if a != '-'])))}
        assert len(ancestry_dict) == k, f'Number of ancestries in training ground truth ({len(ancestry_dict)}) is not equal to the value of K ({k})'
        ancestry_dict['-'] = -1
        to_idx_mapper = np.vectorize(lambda x: ancestry_dict[x])
        # Do not take into account samples with missing labels
        y_num = to_idx_mapper(y[:])
        mask = y_num > -1
        masked_y_num = y_num[mask]
        with dask.config.set(**{'array.slicing.split_large_chunks': False}):
            X_masked = X[mask,:]
        P_init = torch.as_tensor(np.vstack([X_masked[masked_y_num==idx,:].mean(axis=0).compute() for idx in range(k)]), dtype=torch.float32)
        te = time.time()
        log.info('Weights initialized in {} seconds.'.format(te-t0))
        return P_init, None


class PretrainedInitialization(object):
    """Pretrained initialization
    """
    @classmethod
    def get_decoder_init(cls, X, K, path):
        log.info('Fetching pretrained weights...')
        if len(K) > 1:
            raise NotImplementedError("Pretrained mode is only supported for single-head runs.")
        # Loads standard ADMIXTURE output format
        P_init = torch.as_tensor(1-np.genfromtxt(path, delimiter=' ').T, dtype=torch.float32)
        assert P_init.shape[0] == K[0], 'Input P is not coherent with the value of K'
        log.info('Weights fetched.')
        return P_init, None



import argparse
import dask
import dask.array as da
import logging
import pandas as pd
import sys
import torch
import torch.nn as nn
import wandb
from codetiming import Timer
from pathlib import Path
from sklearn.model_selection import KFold
from typing import Dict, List, Tuple, Union

from ..model.neural_admixture import NeuralAdmixture
from ..model.switchers import Switchers
from . import utils

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

def fit_model(trX: da.core.Array, args: argparse.Namespace, valX: Union[None, da.core.Array]=None,
              trY: Union[None, List[str]]=None, valY: Union[None, List[str]]=None,
              dry_run: bool=False) -> Tuple[NeuralAdmixture, torch.device, List[Dict[int, float]]]:
    """Wrapper function to start training

    Args:
        trX (da.core.Array): Dask array containing training data.
        args (argparse.Namespace): parsed argument from CLI.
        valX (Union[None, da.core.Array], optional): Dask array containing validation data. Defaults to None.
        trY (Union[None, List[str]], optional): list containing training labels. Defaults to None.
        valY (Union[None, List[str]], optional): list containing validation labels. Defaults to Union[None, List[str]]=None.
        dry_run (bool, optional): whether to run a dry run (no output is written to disk). Defaults to False.

    Returns:
        Tuple[NeuralAdmixture, torch.device, List[Dict[int, float]]]: instantiated model object along with device and validation error.
    """
    switchers = Switchers.get_switchers()
    num_max_epochs = args.max_epochs
    batch_size = args.batch_size
    learning_rate = args.learning_rate
    save_dir = args.save_dir
    optimizer_str = args.optimizer
    save_every = args.save_every
    l2_penalty = args.l2_penalty
    activation_str = args.activation
    hidden_size = args.hidden_size
    freeze_decoder = bool(args.freeze_decoder)
    init_file = args.init_file
    supervised = bool(args.supervised)
    supervised_loss_weight = float(args.supervised_loss_weight)
    decoder_init = args.initialization if not supervised else 'supervised'
    n_components = int(args.pca_components)
    tol = float(args.tol)
    warmup_epochs = int(args.warmup_epochs)
    name = args.name
    if args.k is not None:
        Ks = [int(args.k)]
    elif args.min_k is not None and args.max_k is not None:
        Ks = [i for i in range(args.min_k, args.max_k+1)]
    else:
        log.error('Either --k (single-head) or --min_k and --max_k (multi-head) must be provided.')
        sys.exit(1)
    assert not (supervised and len(Ks) != 1), 'Supervised version is currently only available on a single head'
    seed = args.seed
    log_to_wandb = bool(args.wandb_log)
    run_name = name
    if not dry_run:
        Path(save_dir).mkdir(parents=True, exist_ok=True)
    log.info(f'Job args: {args}')
    log.info('Will use GPU.' if torch.cuda.is_available() else 'Will use Apple Metal acceleration.' if torch.backends.mps.is_available() else 'No GPUs available.')
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu')
    if log_to_wandb and not dry_run:
        utils.initialize_wandb(run_name, trX, valX, args, save_dir)
    save_path = f'{save_dir}/{run_name}.pt'
    torch.manual_seed(seed)
    # Initialization
    log.info('Initializing...')
    if init_file is None and decoder_init != "pretrained":
        if not dry_run:
            log.warning(f'Initialization filename not provided. Going to store it to {Path(save_dir)/run_name}.pkl')
        init_file = f'{run_name}.pkl'
    init_path = f'{Path(save_dir)/init_file}' if decoder_init != "pretrained" else init_file if not dry_run else None
    P_init, Q_inits = switchers['initializations'][decoder_init](
        trX,
        trY,
        Ks,
        seed,
        init_path,
        run_name,
        n_components,
        batch_size)
    activation = switchers['activations'][activation_str](0)
    log.info(f'Variants: {trX.shape[1]}')
    model = NeuralAdmixture(Ks, trX.shape[1], P_init=P_init,
                                lambda_l2=l2_penalty,
                                encoder_activation=activation,
                                hidden_size=hidden_size,
                                freeze_decoder=freeze_decoder,
                                supervised=supervised,
                                supervised_loss_weight=supervised_loss_weight)
    model.to(device)
    if log_to_wandb:
        wandb.watch(model, log='all', log_freq=1000)

    # Optimizer
    optimizer = switchers['optimizers'][optimizer_str](filter(lambda p: p.requires_grad, model.parameters()), learning_rate)
    log.info('Optimizer successfully loaded.')

    loss_f = nn.BCELoss(reduction='mean')
    log.info('Going to train {} head{}: {}.'.format(len(Ks), 's' if len(Ks) > 1 else '', f'K={Ks[0]}' if len(Ks) == 1 else f'K={Ks[0]} to K={Ks[-1]}'))
    # Fitting
    log.info('Fitting...')
    t = Timer()
    t.start()
    actual_num_epochs = model.launch_training(trX, optimizer, loss_f, num_max_epochs, device, valX=valX,
                       batch_size=batch_size, save_every=save_every,
                       save_path=save_path, trY=trY, valY=valY,
                       seed=seed, log_to_wandb=log_to_wandb, tol=tol,
                       warmup_epochs=warmup_epochs, Q_inits=Q_inits)
    deviances = utils.compute_deviances(model, valX, batch_size, device) if valX is not None else -1
    elapsed_time = t.stop()
    if not dry_run:
        if log_to_wandb:
            wandb.run.summary['total_elapsed_time'] = elapsed_time
            wandb.run.summary['avg_epoch_time'] = elapsed_time/actual_num_epochs
        torch.save(model.state_dict(), save_path)
        model.save_config(run_name, save_dir)
    log.info('Optimization process finished.')
    return model, device, deviances

def main(argv: List[str]):
    """Training entry point
    """
    args = utils.parse_train_args(argv)
    tr_file, val_file = args.data_path, args.validation_data_path
    assert not (val_file and args.cv is not None), 'Cross-validation not available when validation data path is provided.'
    tr_pops_f, val_pops_f = args.populations_path, args.validation_populations_path
    trX, trY, valX, valY = utils.read_data(tr_file, val_file, tr_pops_f, val_pops_f)
    """
    if args.cv is not None:
        log.info(f'Performing {args.cv}-fold cross-validation...')
        cv_obj = KFold(n_splits=args.cv, random_state=args.seed, shuffle=True)
        cv_errs = []
        # TODO: TQDM or improve logging for cross-validation
        for tr_idx, val_idx in cv_obj.split(trX):
            with dask.config.set(**{'array.slicing.split_large_chunks': True}):
                trX_curr, valX_curr = trX[tr_idx], trX[val_idx]
            trY_curr = [trY[idx] for idx in tr_idx] if trY is not None else None
            valY_curr = [trY[idx] for idx in val_idx] if trY is not None else None
            model, device, deviance_curr = fit_model(trX_curr, args, valX_curr, trY_curr, valY_curr, dry_run=True)
            cv_errs += [deviance_curr]

        # TODO: wrap this in a function
        cv_errs = pd.DataFrame.from_records(cv_errs)
        cv_errs_reduced = pd.DataFrame(cv_errs.mean())
        cv_errs_reduced["K"] = cv_errs_reduced.index.copy()
        cv_errs_reduced.rename(columns={0: "cv_error_mean"}, inplace=True)
        cv_errs_reduced["cv_error_std"] = cv_errs.std()
        cv_errs_reduced = cv_errs_reduced.sort_values("K")
    """
    model, device, _ = fit_model(trX, args, valX, trY, valY)
    log.info('Computing divergences...')
    model.display_divergences()
#     if args.cv:
#         for _, row in cv_errs_reduced.iterrows():
#             log.info(f"CV error (K={int(row['K'])}): {row['cv_error_mean']:.5f} ± {row['cv_error_std']:.3f}")
    log.info('Writing outputs...')
    utils.write_outputs(model, trX, valX, args.batch_size, device, args.name, args.save_dir)
    log.info('Exiting...')
    logging.shutdown()
    return 0



# file generated by setuptools_scm
# don't change, don't track in version control
TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Tuple

__version__ = version = '1.3.1'  # type: str
__version_tuple__ = version_tuple = (1, 3, 1)  # type: Tuple[int | str, ...]
